//
// A-synchronous FIFO
//
// Using registers for data storage
//
// Freeware September 2021, Fen Logic Ltd.
// This code is free and is delivered 'as is'.
// It comes without any warranty, to the extent permitted by applicable law.
// There are no restrictions to any use or re-use of this code
// in any form or shape. It would be nice if you keep my company name
// in the source or modified source code but even that is not
// required as I can't check it anyway.
// But the code comes with no warranties or guarantees whatsoever.
//
// Read and Write are ignored when empty/full
//
// Known issues:
//    Not using synchroniser library cells
//

module async_fifo
#(parameter WI  = 16, // data width (in bits of course)
            L2D = 4   // Log2 depth: 4 = 16 deep 
 )
(
   input           reset_n,
   // Write to FIFO interface
   input           w_clk,
   input  [WI-1:0] w_data,
   input           w_strobe,
   output          w_full,
   output  [L2D:0] w_level,
   // Read from FIFO interface
   input           r_clk,
   output [WI-1:0] r_data,
   input           r_strobe,
   output          r_empty,
   output  [L2D:0] r_level
   );

localparam DEPTH = (1<<L2D);

reg  [WI-1:0] memory [0:DEPTH-1];

// For empty/full status we need pointers which are on bit 
// longer then required for memory access.
reg  [L2D:0] write_pnt;       // Write pointer
reg  [L2D:0] rd_meta,rd_sync; // Read pointer from read domain
wire [L2D:0] rd_gray,rd_bin;  // Gray & binary coded read pointer

reg  [L2D:0] read_pnt;        // Read pointer
reg  [L2D:0] wt_meta,wt_sync; // Write pointer from write domain 
wire [L2D:0] wt_gray,wt_bin;  // Gray & binary coded write pointer

// The actual data storage access
wire [L2D-1:0] mem_read,mem_write;


   //
   // All w_clk registers
   // 
   always @(posedge w_clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         write_pnt <= {(L2D+1){1'b0}};
         rd_meta   <= {(L2D+1){1'b0}};
         rd_sync   <= {(L2D+1){1'b0}};
      end
      else
      begin
         // write only if not full 
         if (w_strobe && !w_full)
            write_pnt <= write_pnt + 1;
         
         // The gray code has te be synchronised
         // as only one bit a time will change
         rd_meta <= rd_gray;
         rd_sync <= rd_meta;
      end
   end
   
   // Convert read pointer to gray code
   assign rd_gray = bin2gray(read_pnt);
   
   // Convert synchronised read pointer to binary
   assign rd_bin  = gray2bin(rd_sync);
   
   // For write full status compare write pointer 
   // and synchronised read pointer
   assign w_full  = (rd_bin[L2D]!=write_pnt[L2D]) &&
                    (rd_bin[L2D-1:0]==write_pnt[L2D-1:0]);

   assign w_level =(write_pnt-rd_bin);


   //
   // All r_clk registers
   // 
   always @(posedge r_clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         read_pnt <= {(L2D+1){1'b0}};
         wt_meta  <= {(L2D+1){1'b0}};
         wt_sync  <= {(L2D+1){1'b0}};
      end
      else
      begin
         if (r_strobe && !r_empty)
            read_pnt <= read_pnt + 1;
            
         // The gray code has te be synchronised
         // as only one bit a time will change
         wt_meta <= wt_gray;
         wt_sync <= wt_meta;
      end
   end

   // As with write
   assign wt_gray = bin2gray(write_pnt);
   assign wt_bin  = gray2bin(wt_sync);
   assign r_empty = (wt_bin==read_pnt);

   assign r_level = wt_bin-read_pnt;

//
// Memory
//
   // Strip the MS bit of the pointers
   assign mem_read  = read_pnt[L2D-1:0];
   assign mem_write = write_pnt[L2D-1:0];

   always @(posedge w_clk)
   begin
      if (w_strobe && !w_full)
         memory[mem_write] <= w_data;
   end

   //This is an a-synchronous read!
   assign r_data = memory[mem_read];

//
// Convert binary value to Gray coded value
//
function [L2D:0] bin2gray ;
input  [L2D:0] bin;
begin
   bin2gray = (bin>>1) ^ bin;
end
endfunction

//
// Convert Gray coded value to binary value
//
function [L2D:0] gray2bin;
input  [L2D:0] gray;
reg    [L2D:0] i;
begin
   gray2bin[L2D] = gray[L2D];
   for (i=L2D; i>0; i=i-1)
      gray2bin[i-1] = gray[i-1] ^ gray2bin[i];

end
endfunction

`ifdef ASSERT_ON
   // check basic operating flaws
  
   always @(posedge w_clk)
      if (w_strobe & w_full)
      begin
         $display("%m@%0t: Writing to full FIFO",$time);
         #1 $stop;
      end

   always @(posedge r_clk)
      if (r_strobe & r_empty)
      begin
         $display("%m@%0t: reading from empty FIFO",$time);
         #1 $stop;
      end
  
`endif

endmodule
