//
// Data FIFOs
//
// Freeware September 2015, Fen Logic Ltd.
// This code is free and is delivered 'as is'.
// It comes without any warranty, to the extent permitted by applicable law.
// There are no restrictions to any use or re-use of this code
// in any form or shape. It would be nice if you keep my company name
// in the source or modified source code but even that is not
// required as I can't check it anyway.
// But the code comes with no warranties or guarantees whatsoever.
//
// Minimum register FIFO
// full, empty and level are combinatorial derived
// from read/write pointers: slow!
//
// Known issues:
//    artithmetic is not size compliant
//    Storage contents is not cleared
//


module sync_fifo1
#(parameter WIDTH    = 8, // width in bits
            L2D      = 4  // Log 2 Depth, 4 = 16 deep 
)
(
   input               clk,     // system clock 
   input               reset_n, // A-synchronous low reset/clear
   input               enable,  // clock gating
   input               clear,   // Synchronous clear 

   input               write,   // write FIFO
   input   [WIDTH-1:0] wdata,   // write data
   input               read,    // read FIFO
   output  [WIDTH-1:0] rdata,   // read data

   output              empty,   // FIFO is empty
   output              full,    // FIFO is full
   output [L2D:0] level         // Fill level 

);

localparam DEPTH = (1<<L2D);

reg  [WIDTH-1:0] memory [0:DEPTH-1];

// For empty/full status we need pointers which are on bit 
// bigger then required for memory access.
reg      [L2D:0] rd_pntr,wt_pntr;

wire l_read,l_write;

   // prevent disastrous reads & writes
   assign l_read  = read  & !empty;
   assign l_write = write & !full;


   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         rd_pntr <= {(L2D+1){1'b0}};
         wt_pntr <= {(L2D+1){1'b0}};
      end
      else if (enable)
      begin
         if (clear)
         begin
            rd_pntr <= {(L2D+1){1'b0}};
            wt_pntr <= {(L2D+1){1'b0}};
         end
         else
         begin
           case ({l_read,l_write})
           2'b01 : // Write, no read
             wt_pntr <= wt_pntr + 1;

           2'b10 : // Read, no write
              rd_pntr <= rd_pntr + 1;

           2'b11 : // Read & Write
           begin
              wt_pntr <= wt_pntr + 1;
              rd_pntr <= rd_pntr + 1;
           end
           endcase
        end

`ifdef ASSERT_ON
         // Catch fatal operational errors
         if (write && !l_write)
         begin
            $display("%m,@%0t: write to full FIFO",$time);
            #1 $stop;
         end

         if (read && !l_read)
         begin
            $display("%m,@%0t: read from empty FIFO",$time);
            #1 $stop;
         end
`endif

      end
   end

   assign full  = (rd_pntr[L2D]!=wt_pntr[L2D]) &&
                  (rd_pntr[L2D-1:0]==wt_pntr[L2D-1:0]);
                  
   assign empty = (rd_pntr==wt_pntr);

   assign level = (wt_pntr-rd_pntr);

//
// Memory
//
// The actual data storage access
wire [L2D-1:0] mem_read,mem_write;

   // Strip the MS bit of the pointers
   assign mem_read  = rd_pntr[L2D-1:0];
   assign mem_write = wt_pntr[L2D-1:0];

   always @(posedge clk)
   begin
      if (l_write)
         memory[mem_write] <= wdata;
   end

   //This is an a-synchronous read 
   assign rdata = memory[mem_read];

endmodule